// C Source File
// Created 7/16/2010; 10:35:24 AM

// Delete or comment out the items you do not need.
#define COMMENT_STRING         "Tokenizes Z80 BASIC prgms"
#define COMMENT_PROGRAM_NAME   "Solar89"
#define COMMENT_VERSION_NUMBER 0,0,9,0					/* major, minor, revision, subrevision */
#define COMMENT_AUTHORS        "TC01"
#define COMMENT_BW_ICON \
	{0b0000111111110000, \
	 0b0000000011100000, \
	 0b0000000110000000, \
	 0b0000001100000000, \
	 0b0000111000000000, \
	 0b0001111111111000, \
	 0b0000000000000000, \
	 0b0111110001111100, \
	 0b1100011011000110, \
	 0b1100011011000110, \
	 0b1111111011111110, \
	 0b1111111011111110, \
	 0b1100011000000110, \
	 0b1100011000000110, \
	 0b0111110000000110, \
	 0b0000000000000000}

// Version 0.1 TO-DO lists:
//TO-DO solar89.89z:
//		1. See tokenize.TODO
//TO-DO tokens8x.89p:
//		1. Distr, Format, Zoom, misc tokens
//TO-DO toks.8xt:
//		1. Add all of the TI-83 Plus tokens

#include "solar89.h"

//Makes a dialog to get input strings, returns boolean (integer) of 0 or 1 depending on success
int getDialogInput(char *string)
{
	int result = 0;
	
	HANDLE dlg = H_NULL;

	//Create a dialog box with an "input file" pattern
	if ((dlg = DialogNewSimple(140, 55)) == H_NULL) 
	{
		DlgMessage("DMA Failure", "Unable to create dialog input box", BT_OK, BT_NONE);
		return result;
	}
    
	DialogAddTitle(dlg, "Tokenize Text File", BT_OK, BT_NONE);
	DialogAddRequest(dlg, 5, 15, "Text File:", 0, 17, 18);
	DialogAddRequest(dlg, 5, 25, "Output File:", 17, 17, 18);
	DialogAddRequest(dlg, 5, 35, "Token File:", 34, 17, 18);
	
	//Prompt for input
	if (DialogDo(dlg, CENTER, CENTER, string, NULL) == KEY_ENTER)
	{
		result = 1;
	}

	HeapFree(dlg);
	
	//Return result to say we've done it
	return result;
}

//Extracts the actual program from the text file, returns boolean
int getPrgmFromText(char *filename, char *input, unsigned short size)
{
	FILE *source = NULL;

	int result = 0;
	
	//Open the input file and get its text
	if ((source = fopen(filename, "rt")) == NULL)
	{
		DlgMessage("Open Failure", "Unable to open input file", BT_OK, BT_NONE);
		return result;
	}

	//In theory we should check that fread() works, but for some reason the check doesn't work
	/*if (fread(input, size, 1, source) != 1)
	{
		DlgMessage("Read Failure", "Unable to read input file", BT_OK, BT_NONE);
		return result;
	}*/
	
	result = 1;
	fread(input, size, 1, source);
	fclose(source);
	
	return result;
}

int writePrgmToVar(char *filename, char *output, unsigned short size)
{
	unsigned i;
	int bool = 0;
	unsigned length = 0;

	FILE *result = NULL;

	ST_helpMsg("Writing file");

	//Open a new file for writing in binary mode
	if ((result = fopen(filename, "wb")) == NULL)
	{
		DlgMessage("Open Failure", "Unable to open output file", BT_OK, BT_NONE);
		return bool;
	}
	
	//Calculate length of output file (this causes trouble if we have 0x00 tokens)
	for (i = 0; i < size; ++i)
	{
		if (output[i] != 0)
		{
			length += 1;
		}
	}
	
	//Write the size of the program to the file
	//for now, since we're only testing 1-token projects, this == 1
	//will need a routine for calculating the number of tokens!
	fputc(1, result);
	
	//Write all tokens in the output to file
	for (i = 0; i < length; ++i)
	{
		fputc(output[i], result);		
	}

	//Write the file type
	fputc(0, result);
	fputs("8XP", result);
	fputc(0, result);
	fputc(OTH_TAG, result);

	//Close the file, and if we're successful so far, boolean = true
	fclose(result);
	bool = 1;

	return bool;
}

// Main Function
void _main(void)
{
	char *string = NULL;
	char *inputfile = NULL;
	char *outputfile = NULL;
	char *tokenfile = NULL;
	char *input = NULL;
	unsigned char *output = NULL;

	int bool = 0;
	int length = 0;
	
	SYM_ENTRY *sym;
	unsigned short size = 0;

  //Allocate memory to strings
	if ((string = (char *)calloc(60, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for raw input string", BT_OK, BT_NONE);
		return;
	}
	if ((inputfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for input file string", BT_OK, BT_NONE);
		free(string);
		return;
	}
	if ((outputfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		return;
	}
	if ((tokenfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		return;
	}
	
	//Get user input in string- if we failed to, free memory and quit program
	bool = getDialogInput(string);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		return;
	}

	//Retrieve file names from raw input
	for (length = 0; length < 17; ++length)
	{
		inputfile[length] = string[length];
		outputfile[length] = string[length + 17];
		tokenfile[length] = string[length + 34];
	}
	
	//Find the size of the input and allocate memory accordingly
	sym = SymFindPtr(SYMSTR(inputfile), 0);
	size = ((MULTI_EXPR*)HeapDeref(sym->handle))->Size + 2;
	if ((input = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for input file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		return;
	}

	//Get the raw text of the text file
	bool = getPrgmFromText(inputfile, input, size);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(input);
		return;	
	}
	
	//Allocate memory to the output array, based on input size
	if ((output = (unsigned char *)calloc(size, sizeof(unsigned char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(input);
		return;
	}
	
	//Tokenize (functions in seperate file)
	tokenizePrgm(input, output, tokenfile);

	//Write tokens to application variable
	bool = writePrgmToVar(outputfile, output, size);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(input);
		free(output);
		return;
	}
	
	//Free memory, print msg, and exit
	free(string);
	free(inputfile);
	free(outputfile);
	free(tokenfile);
	free(input);
	free(output);
	ST_helpMsg("Process complete");
}
