// C Source File
// Created 7/17/2010; 12:00:38 PM

// Tokenizer for Solar89
// Converts a text file into an array of hexadecimel tokens

/*TO-DO Here:
		1. Fix address error when trying to free memory to pointers
*/

#include "tokenize.h"

int NUM_TOKENS = 9;

//Initializes all token infos
void initTokens(tokenInfo *tokens, char *tokenfile)
{
	int length;
	int bool;
	int newlength;
	int span = 0;
	int i = 0;
	char newchar;
	char endline = '\r';
	unsigned long rawhex;
	unsigned short size = 0;
	unsigned char hex = '0';

	char *tokentext = NULL;
	char *working = NULL;
	char *line = NULL;
	char *tokenstr = NULL;

	tokenInfo token;
	SYM_ENTRY *sym;
	
	if ((working = (char *)calloc(2, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for working string", BT_OK, BT_NONE);
		return;
	}
	if ((line = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for line string", BT_OK, BT_NONE);
		free(working);
		return;
	}
	if ((tokenstr = (char *)calloc(16, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for token working string", BT_OK, BT_NONE);
		free(working);
		free(line);
		return;
	}
	
	//Get the size of the token file and allocate memory accordingly
	sym = SymFindPtr(SYMSTR(tokenfile), 0);
	size = ((MULTI_EXPR*)HeapDeref(sym->handle))->Size + 2;
	if ((tokentext = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for token file string", BT_OK, BT_NONE);
		free(working);
		free(line);
		free(tokenstr);
		return;
	}
	
	//Get input from file, set length and reallocate memory to troublesome pointers
	bool = getPrgmFromText(tokenfile, tokentext, size);
	if (bool == 0)
	{
		free(tokentext);
		free(working);
		free(line);
		return;
	}
	length = strlen(tokentext);
	line = (char *)realloc(line, (length + 2) * sizeof(char));
	tokenstr = (char *)realloc(tokenstr, (length + 2) * sizeof(char));

	clrscr();
	
	//Loop until the length of the input string is zero
	while (length > 0)
	{
		//Maintain num token size!
		i += 1;
		if (i > NUM_TOKENS)
		{
			tokens = (tokenInfo *)realloc(tokens, i * sizeof(tokenInfo));
			NUM_TOKENS = i;
		}

		//Get the length of the line up to trailing character, reallocate memory
		memset(working, '\0', strlen(working));
		memset(line, '\0', strlen(line));	
		memset(tokenstr, '\0', strlen(tokenstr));			
		sprintf(working, "%c", endline);
		span = strcspn(tokentext, working);
		
		//Load the line into a string, and then remove it
		line = strncat(line, tokentext, span);
		tokentext = strpbrk(tokentext, working);
		
		//Then clear out the \r at the front of the string and clear the working string out
		newchar = tokentext[2];
		sprintf(working, "%c", newchar);
		tokentext = strpbrk(tokentext, working);
		length = strlen(tokentext);
		
		//Init the token
		token = tokens[i];
		token.twoByteToken = 0;

		//Seperate the name component from the line and set it (and it's length!)
		memset(working, '\0', strlen(working));
		sprintf(working, "%c", '\\');
		newlength = strcspn(line, working);
		tokenstr = strncat(tokenstr, line, newlength);
		line += (newlength + 1);
		token.chars = newlength;
		token.name = tokenstr;
		
		//Seperate the hex component from the line and set it
		memset(tokenstr, '\0', strlen(tokenstr));
		tokenstr = strncat(tokenstr, line, 4);
		rawhex = strtol(tokenstr, NULL, 0);
		hex = (unsigned char)rawhex;
		token.hex[0] = hex;
		line += 5;

		//Then check if what's left of the string is the second byte of a two-byte token and deal with it
		if (line[0] != '0')
		{
			rawhex = strtol(tokenstr, NULL, 0);
			ngetchx();
			hex = (unsigned char)rawhex;
			token.hex[1] = hex;
			token.twoByteToken = 1;
		}

		//Update the token
		tokens[i] = token;
	}
	
	//Free memory
	free(working);
	free(tokenstr);
	free(line);
	free(tokentext);
}

//Check if a string passed to it is a token
int isToken(char *string, tokenInfo *tokens)
{
	int i;
	int tokenNum = -1;
	
	tokenInfo token;
	
	//Loop over string tokens
	for (i = 0; i < NUM_TOKENS; ++i)
	{
		token = tokens[i];
		if (strcmp(token.name, string) == 0)
		{
			tokenNum = i;
		}
	}
	
	return tokenNum;
}

//Given a string, this should find the longest token and return its length and hex
tokenInfo findToken(char *string, tokenInfo *tokens)
{
	int i;
	int tokenNum;
	int length = strlen(string);

	tokenInfo token;

	//If the entire thing is a token, deal with it now
	tokenNum = isToken(string, tokens);
	if (tokenNum != -1)
	{
		token = tokens[tokenNum];
		return token;
	}

	//Loop over the characters of the string
	char *truncated = NULL;
	for (i = 0; i < length; ++i)
	{
		//Allocate memory to the truncated string and set it
		if ((truncated = (char *)calloc(length, sizeof(char))) == NULL) 
		{
			DlgMessage("DMA Failure", "Unable to allocate space for truncated string", BT_OK, BT_NONE);
			return token;
		}
		truncated = strncpy(truncated, string, (length-1)-i);
		
		tokenNum = isToken(truncated, tokens);
		if (tokenNum != -1)
		{
			token = tokens[tokenNum];
			free(truncated);
			break;
		}
		
		//Free the truncated string
		free(truncated);
	}

	return token;
}

//Takes a string and a buffer to store generated hex (thanks to calcdude84se!)
void tokenizeString(char *string, unsigned char *buffer, tokenInfo *tokens) 
{
    tokenInfo currentToken;
    while (*string != '\0') 
    {
        currentToken = findToken(string, tokens);
        *(buffer++) = currentToken.hex[0];
        if (currentToken.twoByteToken == 1)
            *(buffer++) = currentToken.hex[1];
        string += currentToken.chars;
    }
}

//Convert a string (from a text file) consisting of 8x tokens in string format, into an array of hexadecimel tokens
void tokenizePrgm(char *input, unsigned char *output, char *tokenfile)
{
	int span = 0;
	int length = strlen(input);
	
	char newchar;
	char endline = '\r';
	char *line = NULL;
	char *working = NULL;
	
	tokenInfo *tokens = NULL;
	
	//Allocate memory to the line string, the working string, and the token array
	if ((line = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for line string", BT_OK, BT_NONE);
		return;
	}
	if ((working = (char *)calloc(2, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for working string", BT_OK, BT_NONE);
		free(line);
		return;
	}
	if ((tokens = (tokenInfo *)calloc(NUM_TOKENS, sizeof(tokenInfo))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for token array", BT_OK, BT_NONE);
		free(line);
		free(working);
		return;
	}

	ST_helpMsg("Initializing tokens");
	
	initTokens(tokens, tokenfile);
	
	ST_helpMsg("Tokenizing");
	
	//Loop until the length of the input string is zero
	while (length > 0)
	{
		//Get the length of the line up to trailing character, reallocate memory
		sprintf(working, "%c", endline);
		span = strcspn(input, working);
		if (line != NULL) 
		{
			line = (char *)realloc(line, span * sizeof(char));
		}
		
		//Load the line into a string, and then remove it
		line = strncpy(line, input, span);
		input = strpbrk(input, working);
		
		//Then clear out the \r at the front of the string and clear the working string out
		newchar = input[2];
		sprintf(working, "%c", newchar);
		input = strpbrk(input, working);
		length = strlen(input);
		
		//Pass the line, the output array, and the token info array to tokenizeString
		tokenizeString(line, output, tokens);
	}
	
	//Free the strings' memory
	free(line);
	free(working);
	
	//Free the token memory
	free(tokens);
}